"""HTTP client."""

from typing import Union

import requests


class HttpClient:
    """HTTP client.

    Will use the API key if the URL starts with the Kili endpoint.
    """

    def __init__(self, kili_endpoint: str, api_key: str, verify: Union[bool, str]) -> None:
        """Initialize the HTTP client."""
        self._kili_endpoint = kili_endpoint.replace("/api/label/v2/graphql", "/api/label/v2")

        self._http_client = requests.Session()
        self._http_client_with_auth = requests.Session()

        self._http_client.verify = verify
        self._http_client_with_auth.verify = verify

        self._auth_headers = {"Authorization": f"X-API-Key: {api_key}"}
        self._http_client_with_auth.headers.update(self._auth_headers)

    def _send_request(self, method: str, url: str, **kwargs) -> requests.Response:
        """Send a request to the given URL."""
        if url.startswith(self._kili_endpoint):
            http_client = self._http_client_with_auth
            # if headers are provided, we need to update the headers with the api key
            if kwargs.get("headers") is not None and "Authorization" not in kwargs["headers"]:
                kwargs["headers"].update(self._auth_headers)
        else:
            http_client = self._http_client

        return http_client.request(method, url, **kwargs)

    def get(self, url: str, **kwargs) -> requests.Response:
        """Send a GET request to the given URL."""
        return self._send_request("GET", url, **kwargs)

    def post(self, url: str, **kwargs) -> requests.Response:
        """Send a POST request to the given URL."""
        return self._send_request("POST", url, **kwargs)

    def head(self, url: str, **kwargs) -> requests.Response:
        """Send a HEAD request to the given URL."""
        return self._send_request("HEAD", url, **kwargs)

    def put(self, url: str, **kwargs) -> requests.Response:
        """Send a PUT request to the given URL."""
        return self._send_request("PUT", url, **kwargs)

    def patch(self, url: str, **kwargs) -> requests.Response:
        """Send a PATCH request to the given URL."""
        return self._send_request("PATCH", url, **kwargs)

    def options(self, url: str, **kwargs) -> requests.Response:
        """Send a OPTIONS request to the given URL."""
        return self._send_request("OPTIONS", url, **kwargs)



"""CLI's project describe command."""

from typing import Optional

import click
from tabulate import tabulate

from kili.domain.project import ProjectId
from kili.entrypoints.cli.common_args import Arguments, Options
from kili.entrypoints.cli.helpers import get_kili_client
from kili.entrypoints.queries.project.helpers import (
    get_project_metadata,
    get_project_metrics,
)


@click.command(name="describe")
@Arguments.project_id
@Options.api_key
@Options.endpoint
def describe_project(api_key: Optional[str], endpoint: Optional[str], project_id: str):
    r"""Show project description and analytics.

    \b
    !!! Examples
        ```
        kili project describe --project-id <project_id>
        ```
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)
    query_fields = [
        "title",
        "id",
        "description",
        "numberOfAssets",
        "numberOfRemainingAssets",
        "numberOfReviewedAssets",
        "numberOfSkippedAssets",
        "honeypotMark",
        "consensusMark",
        "numberOfOpenIssues",
        "numberOfSolvedIssues",
        "numberOfOpenQuestions",
        "numberOfSolvedQuestions",
    ]
    project = kili.kili_api_gateway.get_project(ProjectId(project_id), query_fields)
    metadata = get_project_metadata(project, kili.graphql_client.endpoint)
    dataset_metrics, quality_metrics = get_project_metrics(project)

    print(tabulate(metadata, tablefmt="plain"), end="\n" * 2)
    print("Dataset KPIs", end="\n" + "-" * len("Dataset KPIs") + "\n")
    print(tabulate(dataset_metrics, tablefmt="plain"), end="\n" * 2)
    print("Quality KPIs", end="\n" + "-" * len("Quality KPIs") + "\n")
    print(tabulate(quality_metrics, tablefmt="plain"))



"""Functions to import assets into a TEXT project."""

import json
import os
from enum import Enum
from typing import List, Optional, Tuple

from kili.core.helpers import is_url

from .base import (
    BaseAbstractAssetImporter,
    BatchParams,
    ContentBatchImporter,
)
from .exceptions import ImportFileConversionError, ImportValidationError
from .helpers import is_chat_format, process_json
from .types import AssetLike


class LLMDataType(Enum):
    """LLM data type."""

    DICT = "DICT"
    LIST = "LIST"
    LOCAL_FILE = "LOCAL_FILE"
    HOSTED_FILE = "HOSTED_FILE"


class JSONBatchImporter(ContentBatchImporter):
    """Class for importing a batch of LLM assets with dict content into a LLM_RLHF project."""

    def get_content_type_and_data_from_content(self, content: Optional[str]) -> Tuple[str, str]:
        """Returns the data of the content (path) and its content type."""
        return content or "", "application/json"


class LLMDataImporter(BaseAbstractAssetImporter):
    """Class for importing data into a TEXT project."""

    @staticmethod
    def get_data_type(assets: List[AssetLike]) -> LLMDataType:
        """Determine the type of data to upload from the service payload."""
        content_array = [asset.get("content", None) for asset in assets]
        if all(is_url(content) for content in content_array):
            return LLMDataType.HOSTED_FILE
        if all(isinstance(content, str) and os.path.exists(content) for content in content_array):
            return LLMDataType.LOCAL_FILE
        if all(isinstance(content, dict) for content in content_array):
            return LLMDataType.DICT
        if all(isinstance(content, list) for content in content_array):
            return LLMDataType.LIST
        raise ImportValidationError("Invalid value in content for LLM project.")

    @staticmethod
    def transform_asset_content(asset_content, json_metadata):
        """Transform asset content."""
        content, additional_json_metadata = process_json(asset_content)
        transformed_asset_content = json.dumps(content).encode("utf-8")

        json_metadata_dict = {}
        if json_metadata and isinstance(json_metadata, str):
            json_metadata_dict = json.loads(json_metadata)
        elif json_metadata:
            json_metadata_dict = json_metadata

        merged_json_metadata = {
            **json_metadata_dict,
            **additional_json_metadata,
        }
        changed_json_metadata = json.dumps(merged_json_metadata)

        return transformed_asset_content, changed_json_metadata

    def import_assets(self, assets: List[AssetLike]):
        """Import LLM assets into Kili."""
        self._check_upload_is_allowed(assets)
        data_type = self.get_data_type(assets)
        assets = self.filter_duplicate_external_ids(assets)

        if data_type == LLMDataType.LOCAL_FILE:
            assets = self.filter_local_assets(assets, self.raise_error)
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
            for asset in assets:
                file_path = asset.get("content", None)
                json_metadata = asset.get("json_metadata", "{}")
                if file_path and isinstance(file_path, str):
                    try:
                        with open(file_path, encoding="utf-8") as file:
                            data = json.load(file)

                            if is_chat_format(data, {"role", "content", "id", "chat_id", "model"}):
                                (
                                    asset["content"],
                                    asset["json_metadata"],
                                ) = self.transform_asset_content(data, json_metadata)

                                batch_importer = JSONBatchImporter(
                                    self.kili, self.project_params, batch_params, self.pbar
                                )

                    except Exception as exception:
                        raise ImportFileConversionError(
                            f"Error processing file: {exception}"
                        ) from exception

        elif data_type == LLMDataType.HOSTED_FILE:
            batch_params = BatchParams(is_hosted=True, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        elif data_type in (LLMDataType.DICT, LLMDataType.LIST):
            for asset in assets:
                if "content" in asset and isinstance(asset["content"], dict):
                    asset["content"] = json.dumps(asset["content"]).encode("utf-8")
                elif (
                    "content" in asset
                    and isinstance(asset["content"], list)
                    and is_chat_format(
                        asset["content"], {"role", "content", "id", "chat_id", "model"}
                    )
                ):
                    json_metadata = asset.get("json_metadata", "{}")
                    asset["content"], asset["json_metadata"] = self.transform_asset_content(
                        asset["content"], json_metadata
                    )

            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = JSONBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        else:
            raise ImportValidationError
        return self.import_assets_by_batch(assets, batch_importer)



"""Client presentation methods for cloud storage."""

import logging
from typing import Dict, Generator, Iterable, List, Literal, Optional, overload

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.cloud_storage import (
    DataConnectionFilters,
    DataConnectionId,
    DataIntegrationFilters,
    DataIntegrationId,
    DataIntegrationPlatform,
    DataIntegrationStatus,
    OrganizationId,
    ProjectId,
)
from kili.domain.types import ListOrTuple
from kili.presentation.client.helpers.common_validators import (
    disable_tqdm_if_as_generator,
)
from kili.use_cases.cloud_storage import CloudStorageUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


@for_all_methods(log_call, exclude=["__init__"])
class CloudStorageClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on cloud storage."""

    @overload
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "id",
            "lastChecked",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "id",
            "lastChecked",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "id",
            "lastChecked",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of cloud storage connections that match a set of criteria.

        Args:
            cloud_storage_connection_id: ID of the cloud storage connection.
            cloud_storage_integration_id: ID of the cloud storage integration.
            project_id: ID of the project.
            fields: All the fields to request among the possible fields for the cloud storage connections.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#dataconnection) for all possible fields.
            first: Maximum number of cloud storage connections to return.
            skip: Number of skipped cloud storage connections.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the cloud storage connections is returned.

        Returns:
            A list or a generator of the cloud storage connections that match the criteria.

        Examples:
            >>> kili.cloud_storage_connections(project_id="789465123")
            [{'id': '123456789', 'lastChecked': '2023-02-21T14:49:35.606Z', 'numberOfAssets': 42, 'selectedFolders': ['folder1', 'folder2'], 'projectId': '789465123'}]
        """
        if (
            cloud_storage_connection_id is None
            and cloud_storage_integration_id is None
            and project_id is None
        ):
            raise ValueError(
                "At least one of cloud_storage_connection_id, cloud_storage_integration_id or"
                " project_id must be specified"
            )

        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)

        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        if cloud_storage_connection_id is None:
            data_connections_gen = cloud_storage_use_cases.list_data_connections(
                data_connection_filters=DataConnectionFilters(
                    project_id=ProjectId(project_id) if project_id is not None else None,
                    integration_id=(
                        DataIntegrationId(cloud_storage_integration_id)
                        if cloud_storage_integration_id is not None
                        else None
                    ),
                ),
                fields=fields,
                options=QueryOptions(disable_tqdm, first, skip),
            )
        else:
            data_connections_gen = (
                i
                for i in [
                    cloud_storage_use_cases.get_data_connection(
                        DataConnectionId(cloud_storage_connection_id), fields=fields
                    )
                ]
            )

        if as_generator:
            return data_connections_gen
        return list(data_connections_gen)

    @overload
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("name", "id", "platform", "status"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("name", "id", "platform", "status"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("name", "id", "platform", "status"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of cloud storage integrations that match a set of criteria.

        Args:
            cloud_storage_integration_id: ID of the cloud storage integration.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            organization_id: ID of the organization.
            fields: All the fields to request among the possible fields for the cloud storage integrations.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#dataintegration) for all possible fields.
            first: Maximum number of cloud storage integrations to return.
            skip: Number of skipped cloud storage integrations.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the cloud storage integrations is returned.

        Returns:
            A list or a generator of the cloud storage integrations that match the criteria.

        Examples:
            >>> kili.cloud_storage_integrations()
            [{'name': 'My bucket', 'id': '123456789', 'platform': 'AWS', 'status': 'CONNECTED'}]
        """
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        data_integrations_gen = CloudStorageUseCases(self.kili_api_gateway).list_data_integrations(
            data_integration_filters=DataIntegrationFilters(
                status=status,
                id=(
                    DataIntegrationId(cloud_storage_integration_id)
                    if cloud_storage_integration_id is not None
                    else None
                ),
                name=name,
                platform=platform,
                organization_id=(
                    OrganizationId(organization_id) if organization_id is not None else None
                ),
            ),
            fields=fields,
            options=options,
        )

        if as_generator:
            return data_integrations_gen
        return list(data_integrations_gen)

    @typechecked
    def count_cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
    ) -> int:
        """Count and return the number of cloud storage integrations that match a set of criteria.

        Args:
            cloud_storage_integration_id: ID of the cloud storage integration.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            organization_id: ID of the organization.

        Returns:
            The number of cloud storage integrations that match the criteria.
        """
        return CloudStorageUseCases(self.kili_api_gateway).count_data_integrations(
            DataIntegrationFilters(
                status=status,
                id=(
                    DataIntegrationId(cloud_storage_integration_id)
                    if cloud_storage_integration_id is not None
                    else None
                ),
                name=name,
                platform=platform,
                organization_id=(
                    OrganizationId(organization_id) if organization_id is not None else None
                ),
            )
        )

    @typechecked
    def add_cloud_storage_connection(
        self,
        project_id: str,
        cloud_storage_integration_id: str,
        selected_folders: Optional[List[str]] = None,
    ) -> Dict:
        """Connect a cloud storage to a project.

        Args:
            project_id: Id of the project.
            cloud_storage_integration_id: Id of the cloud storage integration.
            selected_folders: List of folders of the data integration to connect to the project.
                If not provided, all folders of the data integration will be connected.

        Returns:
            A dict with the DataConnection Id.
        """
        data_connection_id = CloudStorageUseCases(self.kili_api_gateway).add_data_connection(
            project_id=ProjectId(project_id),
            data_integration_id=DataIntegrationId(cloud_storage_integration_id),
            selected_folders=selected_folders,
            fields=("id",),
        )["id"]

        return {"id": data_connection_id}

    @typechecked
    def synchronize_cloud_storage_connection(
        self,
        cloud_storage_connection_id: str,
        delete_extraneous_files: bool = False,
        dry_run: bool = False,
    ) -> Dict:
        """Synchronize a cloud storage connection.

        This method will compute differences between the cloud storage connection and the project,
            and then validate the differences.

        If `delete_extraneous_files` is True, it will also delete files that are not in the
            cloud storage integration anymore but that are still in the project.

        Args:
            cloud_storage_connection_id: Id of the cloud storage connection.
            delete_extraneous_files: If True, delete extraneous files.
            dry_run: If True, will not synchronize the data connection but only print the
                differences. This is useful to check the differences before applying them to the
                project.

        Returns:
            A dict with the cloud storage connection Id.
        """
        data_connection_id = DataConnectionId(cloud_storage_connection_id)

        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        cloud_storage_use_cases.synchronize_data_connection(
            data_connection_id=data_connection_id,
            delete_extraneous_files=delete_extraneous_files,
            dry_run=dry_run,
            logger=logger,
        )

        return cloud_storage_use_cases.get_data_connection(
            data_connection_id=data_connection_id, fields=("numberOfAssets", "projectId")
        )

    @typechecked
    def create_cloud_storage_integration(
        self,
        platform: DataIntegrationPlatform,
        name: str,
        fields: ListOrTuple[str] = (
            "id",
            "name",
            "status",
            "platform",
            "allowedPaths",
        ),
        allowed_paths: Optional[List[str]] = None,
        allowed_projects: Optional[List[str]] = None,
        aws_access_point_arn: Optional[str] = None,
        aws_role_arn: Optional[str] = None,
        aws_role_external_id: Optional[str] = None,
        azure_connection_url: Optional[str] = None,
        azure_is_using_service_credentials: Optional[bool] = None,
        azure_sas_token: Optional[str] = None,
        azure_tenant_id: Optional[str] = None,
        gcp_bucket_name: Optional[str] = None,
        include_root_files: Optional[str] = None,
        internal_processing_authorized: Optional[str] = None,
        s3_access_key: Optional[str] = None,
        s3_bucket_name: Optional[str] = None,
        s3_endpoint: Optional[str] = None,
        s3_region: Optional[str] = None,
        s3_secret_key: Optional[str] = None,
        s3_session_token: Optional[str] = None,
    ) -> Dict:
        # pylint: disable=line-too-long
        """Create a cloud storage integration.

        Args:
            fields: All the fields to request among the possible fields for the cloud storage integration.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#dataintegration) for all possible fields.
            allowed_paths: List of allowed paths.
            allowed_projects: List of allowed projects.
            aws_access_point_arn: AWS access point ARN.
            aws_role_arn: AWS role ARN.
            aws_role_external_id: AWS role external ID.
            azure_connection_url: Azure connection URL.
            azure_is_using_service_credentials: Whether Azure is using service credentials.
            azure_sas_token: Azure SAS token.
            azure_tenant_id: Azure tenant ID.
            gcp_bucket_name: GCP bucket name.
            include_root_files: Whether to include root files.
            internal_processing_authorized: Whether internal processing is authorized.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            s3_access_key: S3 access key.
            s3_bucket_name: S3 bucket name.
            s3_endpoint: S3 endpoint.
            s3_region: S3 region.
            s3_secret_key: S3 secret key.
            s3_session_token: S3 session token.
        """
        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        return cloud_storage_use_cases.create_data_integration(
            platform=platform,
            name=name,
            fields=fields,
            allowed_paths=allowed_paths,
            allowed_projects=allowed_projects,
            aws_access_point_arn=aws_access_point_arn,
            aws_role_arn=aws_role_arn,
            aws_role_external_id=aws_role_external_id,
            azure_connection_url=azure_connection_url,
            azure_is_using_service_credentials=azure_is_using_service_credentials,
            azure_sas_token=azure_sas_token,
            azure_tenant_id=azure_tenant_id,
            gcp_bucket_name=gcp_bucket_name,
            include_root_files=include_root_files,
            internal_processing_authorized=internal_processing_authorized,
            s3_access_key=s3_access_key,
            s3_bucket_name=s3_bucket_name,
            s3_endpoint=s3_endpoint,
            s3_region=s3_region,
            s3_secret_key=s3_secret_key,
            s3_session_token=s3_session_token,
        )

    @typechecked
    def update_cloud_storage_integration(
        self,
        cloud_storage_integration_id: str,
        allowed_paths: Optional[List[str]] = None,
        allowed_projects: Optional[List[str]] = None,
        aws_access_point_arn: Optional[str] = None,
        aws_role_arn: Optional[str] = None,
        aws_role_external_id: Optional[str] = None,
        azure_connection_url: Optional[str] = None,
        azure_is_using_service_credentials: Optional[bool] = None,
        azure_sas_token: Optional[str] = None,
        azure_tenant_id: Optional[str] = None,
        gcp_bucket_name: Optional[str] = None,
        include_root_files: Optional[str] = None,
        internal_processing_authorized: Optional[str] = None,
        name: Optional[str] = None,
        organization_id: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        s3_access_key: Optional[str] = None,
        s3_bucket_name: Optional[str] = None,
        s3_endpoint: Optional[str] = None,
        s3_region: Optional[str] = None,
        s3_secret_key: Optional[str] = None,
        s3_session_token: Optional[str] = None,
    ) -> Dict:
        """Update cloud storage data integration.

        Args:
            allowed_paths: List of allowed paths.
            allowed_projects: List of allowed projects.
            aws_access_point_arn: AWS access point ARN.
            aws_role_arn: AWS role ARN.
            aws_role_external_id: AWS role external ID.
            azure_connection_url: Azure connection URL.
            azure_is_using_service_credentials: Whether Azure is using service credentials.
            azure_sas_token: Azure SAS token.
            azure_tenant_id: Azure tenant ID.
            cloud_storage_integration_id: Data integration ID.
            gcp_bucket_name: GCP bucket name.
            include_root_files: Whether to include root files.
            internal_processing_authorized: Whether internal processing is authorized.
            organization_id: Organization ID.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            s3_access_key: S3 access key.
            s3_bucket_name: S3 bucket name.
            s3_endpoint: S3 endpoint.
            s3_region: S3 region.
            s3_secret_key: S3 secret key.
            s3_session_token: S3 session token.
        """
        return CloudStorageUseCases(self.kili_api_gateway).update_data_integration(
            data_integration_id=DataIntegrationId(cloud_storage_integration_id),
            name=name,
            platform=platform,
            allowed_paths=allowed_paths,
            allowed_projects=allowed_projects,
            aws_access_point_arn=aws_access_point_arn,
            aws_role_arn=aws_role_arn,
            aws_role_external_id=aws_role_external_id,
            azure_connection_url=azure_connection_url,
            azure_is_using_service_credentials=azure_is_using_service_credentials,
            azure_sas_token=azure_sas_token,
            azure_tenant_id=azure_tenant_id,
            gcp_bucket_name=gcp_bucket_name,
            include_root_files=include_root_files,
            internal_processing_authorized=internal_processing_authorized,
            organization_id=organization_id,
            s3_access_key=s3_access_key,
            s3_bucket_name=s3_bucket_name,
            s3_endpoint=s3_endpoint,
            s3_region=s3_region,
            s3_secret_key=s3_secret_key,
            s3_session_token=s3_session_token,
            status=status,
        )

    @typechecked
    def delete_cloud_storage_integration(self, cloud_storage_integration_id: str) -> str:
        """Delete a cloud storage integration.

        Args:
            cloud_storage_integration_id: Id of the cloud storage integration.
        """
        cloud_storage_integration_id = DataIntegrationId(cloud_storage_integration_id)

        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        return cloud_storage_use_cases.delete_data_integration(
            data_integration_id=cloud_storage_integration_id
        )



"""Set of utils for label data parsing service module."""

from typing import Dict, List


def get_children_job_names(json_interface: Dict, job_interface: Dict) -> List[str]:
    """Returns the list of children job names of a parent job interface."""
    children_job_names = []

    for key, value in job_interface.items():
        if (
            key == "children"
            and isinstance(value, List)
            and all(job_name in json_interface for job_name in value)
            and all(json_interface[job_name]["isChild"] for job_name in value)
        ):
            children_job_names.extend(value)

        elif isinstance(value, Dict):
            children_job_names.extend(
                get_children_job_names(json_interface=json_interface, job_interface=value)
            )

        # pose estimation json interface
        elif key == "points" and isinstance(value, List):
            for point in value:
                children_job_names.extend(
                    get_children_job_names(json_interface=json_interface, job_interface=point)
                )

    return children_job_names



"""Module for the "categories" key parsing of a job response."""

from typing import Any, Dict, Iterator, List, Optional

from typeguard import typechecked

import kili.services.label_data_parsing.json_response as json_response_module

from .exceptions import InvalidMutationError
from .types import Project
from .utils import get_children_job_names


class Category:
    """Class for Category parsing."""

    def __init__(self, category_json: Dict, project_info: Project, job_name: str) -> None:
        """Class for Category parsing.

        Args:
            category_json: Value of the key "categories".
            project_info: Information about the project.
            job_name: Name of the job.
        """
        self._json_data = {}
        self._project_info = project_info
        self._job_name = job_name
        self._job_interface = project_info["jsonInterface"][job_name]  # type: ignore
        # call the setters to check the values are valid
        self.name = category_json["name"]
        if "confidence" in category_json:
            self.confidence = category_json["confidence"]
        if category_json.get("children"):
            self.children = category_json["children"]

    def __str__(self) -> str:
        """Return the string representation of the category."""
        return str(self._json_data)

    def __repr__(self) -> str:
        """Return the string representation of the category."""
        return repr(self._json_data)

    def as_dict(self) -> Dict:
        """Return the parsed category as a dict."""
        ret = {"name": self._json_data["name"]}
        if "confidence" in self._json_data:
            ret["confidence"] = self._json_data["confidence"]
        if "children" in self._json_data:
            ret["children"] = (
                self._json_data["children"].to_dict()
                if not isinstance(self._json_data["children"], Dict)
                else self._json_data["children"]
            )
        return ret

    @property
    def name(self) -> str:
        """Return the name of the category label.

        For a json interface such as:

        ```json
        "CLASSIFICATION_JOB": {
            "mlTask": "CLASSIFICATION",
            "content": {
                "categories": {
                    "CATEGORY_A": {"name": "Category A"},
                    "CATEGORY_B": {"name": "Category B"},
                },
            },
        }
        ```

        The name of the category label can be `CATEGORY_A` or `CATEGORY_B`.

        To get the displayed name of the category label (`Category A`, or `Category B`),
            use the `.display_name` attribute instead.
        """
        return self._json_data["name"]

    @property
    def display_name(self) -> str:
        """Return the displayed name of the category label.

        It is the category name that is displayed in the UI.

        For a json interface such as:

        ```json
        "CLASSIFICATION_JOB": {
            "mlTask": "CLASSIFICATION",
            "content": {
                "categories": {
                    "CATEGORY_A": {"name": "Category A"},
                    "CATEGORY_B": {"name": "Category B"},
                },
            },
        }
        ```

        The displayed name of the category label can be `Category A` or `Category B`.

        To get the name of the category label (`CATEGORY_A` or `CATEGORY_B`),
            use the `.name` attribute instead.
        """
        return self._json_data["display_name"]

    @name.setter
    @typechecked
    def name(self, name: str) -> None:
        """Set the name of the category label."""
        for cat_key, cat_val in self._job_interface["content"]["categories"].items():
            if name == cat_key:
                self._json_data["name"] = name
                self._json_data["display_name"] = cat_val["name"]
                return

        raise InvalidMutationError(
            f"Category '{name}' is not in the job interface with categories:"
            f" {list(self._job_interface['content']['categories'].keys())}"
        )

    @display_name.setter
    @typechecked
    def display_name(self, name: str) -> None:
        """Set the displayed name of the category label."""
        ui_name_to_name = {}
        for cat_key, cat_val in self._job_interface["content"]["categories"].items():
            if cat_val["name"] in ui_name_to_name:
                raise InvalidMutationError(
                    f"Category '{cat_val}' is not unique in the job interface:"
                    f" {self._job_interface}\nUse `.name` instead of `.display_name` to set the"
                    " category name.`"
                )
            ui_name_to_name[cat_val["name"]] = cat_key

        try:
            self._json_data["name"] = ui_name_to_name[name]
            self._json_data["display_name"] = name
        except KeyError as err:
            raise InvalidMutationError(
                f"Category '{name}' is not in the job interface with categories:"
                f" {ui_name_to_name.keys()}"
            ) from err

    @property
    def confidence(self) -> float:
        """Returns the confidence of the category label."""
        return self._json_data["confidence"]

    @confidence.setter
    @typechecked
    def confidence(self, confidence: float) -> None:
        """Set the confidence of the category label."""
        self._json_data["confidence"] = confidence

    @property
    def children(self) -> "json_response_module.ParsedJobs":
        """Returns the parsed children jobs of the classification job."""
        return self._json_data["children"]

    @children.setter
    @typechecked
    def children(self, children: Dict) -> None:
        """Set the children jobs of the classification job."""
        job_names_to_parse = get_children_job_names(
            json_interface=self._project_info["jsonInterface"],
            job_interface=self._job_interface,  # pyright: ignore [reportGeneralTypeIssues]
        )
        parsed_children_job = json_response_module.ParsedJobs(
            project_info=self._project_info,
            json_response=children,
            job_names_to_parse=job_names_to_parse,
        )
        self._json_data["children"] = parsed_children_job


class CategoryList:
    """Class for the categories list parsing."""

    def __init__(
        self, categories_list: List[Dict[str, Any]], project_info: Project, job_name: str
    ) -> None:
        """Class for the categories list parsing.

        Args:
            categories_list: List of dicts representing categories.
            project_info: Information about the project.
            job_name: Name of the job.
        """
        self._categories_list: List[Category] = []
        self._project_info = project_info
        self._job_name = job_name

        self._job_interface = project_info["jsonInterface"][job_name]  # type: ignore

        for category_dict in categories_list:
            self.add_category(**category_dict)

    def _check_can_append_category(self, category: Category) -> None:
        input_type = self._job_interface["content"]["input"]
        nb_classes = len(self._job_interface["content"]["categories"])
        len_categories = len(self._categories_list)

        if input_type in {"radio", "singleDropdown"}:
            if len_categories >= 1:
                raise InvalidMutationError(
                    f"Cannot add more than one category to a {input_type} classification job."
                )

        elif input_type in {"checkbox", "multipleDropdown"} and len_categories >= nb_classes:
            raise InvalidMutationError(
                f"Cannot add more categories than the number of classes to a {input_type}"
                " classification job."
            )

        # Check that the name of the category we want to add is not already in the list
        if any(category.name == cat.name for cat in self._categories_list):
            raise InvalidMutationError(
                f"Cannot add a category with name '{category.name}' because a category with the"
                f" same name already exists: {self._categories_list}"
            )

    @typechecked
    def add_category(
        self,
        name: str,
        confidence: Optional[float] = None,
        children: Optional[Dict] = None,
    ) -> None:
        """Add a category object to the CategoryList object."""
        category_dict: Dict[str, object] = {"name": name}
        if confidence is not None:
            category_dict["confidence"] = confidence
        if children is not None:
            category_dict["children"] = children

        category = Category(
            category_json=category_dict,
            project_info=self._project_info,
            job_name=self._job_name,
        )
        self._check_can_append_category(category)
        self._categories_list.append(category)

    @typechecked
    def __getitem__(self, index: int) -> Category:
        """Return the category object at the given index."""
        return self._categories_list[index]

    def __len__(self) -> int:
        """Return the number of categories."""
        return len(self._categories_list)

    def __iter__(self) -> Iterator[Category]:
        """Return an iterator over the categories."""
        return iter(self._categories_list)

    def __str__(self) -> str:
        """Return the string representation of the categories list."""
        return str(self.as_list())

    def __repr__(self) -> str:
        """Return the string representation of the categories list."""
        return repr(self.as_list())

    def as_list(self) -> List[Dict]:
        """Return the list of categories as a list of dicts."""
        return [category.as_dict() for category in self._categories_list]



"""Issue domain."""

from dataclasses import dataclass
from typing import List, Literal, NewType, Optional

from kili.domain.project import ProjectId

IssueType = Literal["ISSUE", "QUESTION"]
IssueStatus = Literal["OPEN", "SOLVED"]
IssueId = NewType("IssueId", str)


@dataclass
class IssueFilters:
    """Issue filters for running an issue search."""

    project_id: ProjectId
    asset_id: Optional[str] = None
    asset_id_in: Optional[List[str]] = None
    issue_type: Optional[IssueType] = None
    status: Optional[IssueStatus] = None



"""Client presentation methods for users."""

from typing import Dict, Generator, Iterable, List, Literal, Optional, overload

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.enums import OrganizationRole
from kili.domain.organization import OrganizationId
from kili.domain.types import ListOrTuple
from kili.domain.user import UserFilter
from kili.presentation.client.helpers.common_validators import (
    disable_tqdm_if_as_generator,
)
from kili.use_cases.user import UserUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods


@for_all_methods(log_call, exclude=["__init__"])
class UserClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on users."""

    @overload
    def users(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("email", "id", "firstname", "lastname"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def users(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("email", "id", "firstname", "lastname"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def users(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("email", "id", "firstname", "lastname"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of users given a set of criteria.

        Args:
            api_key: Query an user by its API Key
            email: Email of the user
            organization_id: Identifier of the user's organization
            fields: All the fields to request among the possible fields for the users.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#user) for all possible fields.
            first: Maximum number of users to return
            skip: Number of skipped users (they are ordered by creation date)
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the users is returned.

        Returns:
            An iterable of users.

        Examples:
            ```
            # List all users in my organization
            >>> organization = kili.organizations()[0]
            >>> organization_id = organization['id']
            >>> kili.users(organization_id=organization_id)
            ```
        """
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)

        users_gen = UserUseCases(self.kili_api_gateway).list_users(
            filters=UserFilter(
                api_key=api_key,
                email=email,
                organization_id=OrganizationId(organization_id) if organization_id else None,
                activated=None,
                id=None,
                id_in=None,
            ),
            fields=fields,
            options=QueryOptions(disable_tqdm, first, skip),
        )

        if as_generator:
            return users_gen
        return list(users_gen)

    @typechecked
    def count_users(
        self,
        organization_id: Optional[str] = None,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
    ) -> int:
        """Get user count based on a set of constraints.

        Args:
            organization_id: Identifier of the user's organization.
            api_key: Filter by API Key.
            email: Filter by email.

        Returns:
            The number of organizations with the parameters provided.
        """
        return UserUseCases(self.kili_api_gateway).count_users(
            UserFilter(
                api_key=api_key,
                email=email,
                organization_id=OrganizationId(organization_id) if organization_id else None,
                activated=None,
                id=None,
                id_in=None,
            )
        )

    @typechecked
    def create_user(
        self,
        email: str,
        password: str,
        organization_role: OrganizationRole,
        firstname: Optional[str] = None,
        lastname: Optional[str] = None,
    ) -> Dict[Literal["id"], str]:
        """Add a user to your organization.

        Args:
            email: Email of the new user, used as user's unique identifier.
            password: On the first sign in, he will use this password and be able to change it.
            organization_role: One of "ADMIN", "USER".
            firstname: First name of the new user.
            lastname: Last name of the new user.

        Returns:
            A dictionary with the id of the new user.
        """
        return UserUseCases(self.kili_api_gateway).create_user(
            email=email.lower(),
            password=password,
            organization_role=organization_role,
            firstname=firstname,
            lastname=lastname,
            fields=("id",),
        )

    @typechecked
    def update_password(
        self, email: str, old_password: str, new_password_1: str, new_password_2: str
    ) -> Dict[Literal["id"], str]:
        """Allow to modify the password that you use to connect to Kili.

        This resolver only works for on-premise installations without Auth0.

        Args:
            email: Email of the person whose password has to be updated.
            old_password: The old password
            new_password_1: The new password
            new_password_2: A confirmation field for the new password

        Returns:
            A dict with the user id.
        """
        return UserUseCases(self.kili_api_gateway).update_password(
            old_password=old_password,
            new_password_1=new_password_1,
            new_password_2=new_password_2,
            user_filter=UserFilter(
                email=email,
                activated=None,
                api_key=None,
                id=None,
                id_in=None,
                organization_id=None,
            ),
            fields=("id",),
        )

    @typechecked
    def update_properties_in_user(
        self,
        email: str,
        firstname: Optional[str] = None,
        lastname: Optional[str] = None,
        organization_id: Optional[str] = None,
        organization_role: Optional[OrganizationRole] = None,
        activated: Optional[bool] = None,
    ) -> Dict[Literal["id"], str]:
        """Update the properties of a user.

        Args:
            email: The email is the identifier of the user.
            firstname: Change the first name of the user.
            lastname: Change the last name of the user.
            organization_id: Change the organization the user is related to.
            organization_role: Change the role of the user.
                One of "ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER".
            activated: In case we want to deactivate a user, but keep it.

        Returns:
            A dict with the user id.
        """
        return UserUseCases(self.kili_api_gateway).update_user(
            user_filter=UserFilter(
                email=email,
                activated=None,
                api_key=None,
                id=None,
                id_in=None,
                organization_id=None,
            ),
            firstname=firstname,
            lastname=lastname,
            organization_id=OrganizationId(organization_id) if organization_id else None,
            organization_role=organization_role,
            activated=activated,
            fields=("id",),
        )



"""Cloud Storage domain."""

from dataclasses import dataclass
from enum import Enum
from typing import Literal, NewType, Optional

from .organization import OrganizationId
from .project import ProjectId

DataIntegrationId = NewType("DataIntegrationId", str)
DataConnectionId = NewType("DataConnectionId", str)

DataIntegrationPlatform = Literal["AWS", "Azure", "GCP", "CustomS3"]

DataIntegrationStatus = Literal["CONNECTED", "DISCONNECTED", "CHECKING"]


class DataDifferenceType(str, Enum):
    """Data difference type."""

    ADD = "ADD"
    REMOVE = "REMOVE"


@dataclass
class DataIntegrationFilters:
    """Data Integration filters."""

    id: Optional[DataIntegrationId]
    status: Optional[DataIntegrationStatus] = None
    name: Optional[str] = None
    organization_id: Optional[OrganizationId] = None
    platform: Optional[DataIntegrationPlatform] = None


@dataclass
class DataConnectionFilters:
    """DataConnectionsWhere filters."""

    project_id: Optional[ProjectId]
    integration_id: Optional[DataIntegrationId]



"""Exceptions for label response module."""


class JobNotExistingError(Exception):
    """Raised when the job does not exist."""

    def __init__(self, job_name: str) -> None:
        """Init."""
        super().__init__(f"Job named '{job_name}' does not exist.")


class AttributeNotCompatibleWithJobError(Exception):
    """Raised when the attribute is not compatible with the job."""

    def __init__(self, attribute_name: str) -> None:
        """Init."""
        super().__init__(f"The attribute '{attribute_name}' is not compatible with the job.")


class InvalidMutationError(Exception):
    """Raised when the mutation is invalid."""


class FrameIndexError(Exception):
    """Raised when the frame index is invalid."""

    def __init__(self, frame_index: int, nb_frames: int) -> None:
        """Init."""
        super().__init__(
            f"Frame index {frame_index} out of range for frame list of size {nb_frames}."
        )



"""Project use cases."""

import json
from typing import Dict, Generator, Optional

from tenacity import Retrying
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_fixed

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.project.mappers import project_data_mapper
from kili.adapters.kili_api_gateway.project.types import ProjectDataKiliAPIGatewayInput
from kili.core.enums import ProjectType
from kili.domain.project import ComplianceTag, InputType, ProjectFilters, ProjectId
from kili.domain.types import ListOrTuple
from kili.exceptions import NotFound
from kili.use_cases.base import BaseUseCases


class ProjectUseCases(BaseUseCases):
    """Project use cases."""

    # pylint: disable=too-many-arguments
    def create_project(
        self,
        title: str,
        description: str,
        project_type: Optional[ProjectType],
        compliance_tags: Optional[ListOrTuple[ComplianceTag]],
        project_id: Optional[ProjectId] = None,
        input_type: Optional[InputType] = None,
        json_interface: Optional[Dict] = None,
    ) -> ProjectId:
        """Create or copy a project if project_id is set."""
        if project_id is not None:
            project_copied = self._kili_api_gateway.get_project(
                project_id=project_id, fields=["jsonInterface", "instructions", "inputType"]
            )
            project_tag = self._kili_api_gateway.list_tags_by_project(
                project_id=project_id, fields=["id"]
            )
            new_project_id = self._kili_api_gateway.create_project(
                input_type=project_copied["inputType"],
                json_interface=project_copied["jsonInterface"],
                title=title,
                description=description,
                project_type=project_type,
                compliance_tags=compliance_tags,
            )
            if project_copied["instructions"]:
                self.update_properties_in_project(
                    project_id=new_project_id,
                    instructions=project_copied["instructions"],
                )
            tags_of_orga = self._kili_api_gateway.list_tags_by_org(fields=("id",))
            tags_of_orga_ids = [tag["id"] for tag in tags_of_orga]

            for tag in project_tag:
                if tag["id"] not in tags_of_orga_ids:
                    raise ValueError(
                        f"Tag {tag['id']} doesn't belong to your organization and was not copied."
                    )
                self._kili_api_gateway.check_tag(project_id=new_project_id, tag_id=tag["id"])
        elif input_type is None or json_interface is None:
            raise ValueError(
                """Arguments `input_type` and `json_interface` must be set
                if no `project_id` is providen."""
            )
        else:
            new_project_id = self._kili_api_gateway.create_project(
                input_type=input_type,
                json_interface=json_interface,
                title=title,
                description=description,
                project_type=project_type,
                compliance_tags=compliance_tags,
            )

        # The project is not immediately available after creation
        for attempt in Retrying(
            stop=stop_after_delay(60),
            wait=wait_fixed(1),
            retry=retry_if_exception_type(NotFound),
            reraise=True,
        ):
            with attempt:
                _ = self._kili_api_gateway.get_project(project_id=new_project_id, fields=("id",))

        return ProjectId(new_project_id)

    def list_projects(
        self,
        project_filters: ProjectFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """Return a generator of projects that match the filter."""
        return self._kili_api_gateway.list_projects(project_filters, fields, options=options)

    def count_projects(self, project_filters: ProjectFilters) -> int:
        """Return the number of projects that match the filter."""
        return self._kili_api_gateway.count_projects(project_filters)

    # pylint: disable=too-many-locals
    def update_properties_in_project(
        self,
        project_id: ProjectId,
        *,
        can_navigate_between_assets: Optional[bool] = None,
        can_skip_asset: Optional[bool] = None,
        compliance_tags: Optional[ListOrTuple[ComplianceTag]] = None,
        consensus_mark: Optional[float] = None,
        consensus_tot_coverage: Optional[int] = None,
        description: Optional[str] = None,
        honeypot_mark: Optional[float] = None,
        instructions: Optional[str] = None,
        input_type: Optional[InputType] = None,
        json_interface: Optional[Dict] = None,
        min_consensus_size: Optional[int] = None,
        number_of_assets: Optional[int] = None,
        number_of_skipped_assets: Optional[int] = None,
        number_of_remaining_assets: Optional[int] = None,
        number_of_reviewed_assets: Optional[int] = None,
        review_coverage: Optional[int] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        title: Optional[str] = None,
        use_honeypot: Optional[bool] = None,
        metadata_types: Optional[Dict] = None,
    ) -> Dict[str, object]:
        """Update properties in a project."""
        if consensus_tot_coverage is not None and not 0 <= consensus_tot_coverage <= 100:
            raise ValueError(
                "Argument `consensus_tot_coverage` must be comprised between 0 and 100."
            )

        if min_consensus_size is not None and not 1 <= min_consensus_size <= 10:
            raise ValueError("Argument `min_consensus_size` must be comprised between 1 and 10.")

        if review_coverage is not None and not 0 <= review_coverage <= 100:
            raise ValueError("Argument `review_coverage` must be comprised between 0 and 100.")

        project_data = ProjectDataKiliAPIGatewayInput(
            can_navigate_between_assets=can_navigate_between_assets,
            can_skip_asset=can_skip_asset,
            consensus_mark=consensus_mark,
            consensus_tot_coverage=consensus_tot_coverage,
            compliance_tags=compliance_tags,
            description=description,
            honeypot_mark=honeypot_mark,
            instructions=instructions,
            input_type=input_type,
            json_interface=json.dumps(json_interface) if json_interface is not None else None,
            metadata_types=metadata_types,
            min_consensus_size=min_consensus_size,
            number_of_assets=number_of_assets,
            number_of_skipped_assets=number_of_skipped_assets,
            number_of_remaining_assets=number_of_remaining_assets,
            number_of_reviewed_assets=number_of_reviewed_assets,
            review_coverage=review_coverage,
            should_relaunch_kpi_computation=should_relaunch_kpi_computation,
            title=title,
            use_honeypot=use_honeypot,
            archived=None,
            author=None,
            rules=None,
        )

        fields = tuple(
            name for name, val in project_data_mapper(project_data).items() if val is not None
        )
        if "id" not in fields:
            fields += ("id",)

        return self._kili_api_gateway.update_properties_in_project(project_id, project_data, fields)



"""Mixin extending Kili API Gateway class with Organization related operations."""

from typing import Dict, Generator

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.adapters.kili_api_gateway.organization.mappers import (
    map_create_organization_data,
    map_list_organizations_where,
    map_organization_metrics_where,
    map_update_organization_data,
    map_update_organization_where,
)
from kili.domain.organization import (
    OrganizationFilters,
    OrganizationId,
    OrganizationMetricsFilters,
    OrganizationToCreateInput,
    OrganizationToUpdateInput,
)
from kili.domain.types import ListOrTuple

from .operations import (
    COUNT_ORGANIZATIONS_QUERY,
    get_create_organization_mutation,
    get_get_organization_metrics_query,
    get_list_organizations_query,
    get_update_properties_in_organization_mutation,
)


class OrganizationOperationMixin(BaseOperationMixin):
    """GraphQL Mixin extending GraphQL Gateway class with Organization related operations."""

    def create_organization(
        self,
        organization: OrganizationToCreateInput,
    ) -> Dict:
        """Send a GraphQL request calling createOrganization resolver."""
        payload = {"data": map_create_organization_data(organization)}
        fragment = fragment_builder(["id"])
        result = self.graphql_client.execute(get_create_organization_mutation(fragment), payload)

        return result["data"]

    def list_organizations(
        self,
        filters: OrganizationFilters,
        fields: ListOrTuple[str],
        description: str,
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """Send a series of GraphQL request calling organizations resolver."""
        fragment = fragment_builder(fields)
        query = get_list_organizations_query(fragment)
        where = map_list_organizations_where(filters=filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, description, COUNT_ORGANIZATIONS_QUERY
        )

    def count_organizations(self, filters: OrganizationFilters) -> int:
        """Send a GraphQL request calling countOrganizations resolver."""
        where = map_list_organizations_where(filters=filters)
        payload = {"where": where}
        count_result = self.graphql_client.execute(COUNT_ORGANIZATIONS_QUERY, payload)
        count: int = count_result["data"]
        return count

    def get_organization_metrics(
        self, filters: OrganizationMetricsFilters, fields: ListOrTuple[str]
    ) -> Dict:
        """Send a GraphQL request calling organizationMetrics resolver."""
        where = map_organization_metrics_where(filters=filters)
        payload = {"where": where}
        fragment = fragment_builder(fields)
        result = self.graphql_client.execute(get_get_organization_metrics_query(fragment), payload)
        return result["data"]

    def update_organization(
        self,
        organization_id: OrganizationId,
        organization_data: OrganizationToUpdateInput,
    ) -> Dict:
        """Send a GraphQL request calling updateOrganization resolver."""
        data = map_update_organization_data(organization_data)
        where = map_update_organization_where(organization_id)
        payload = {"where": where, "data": data}
        fragment = fragment_builder(["id"])
        result = self.graphql_client.execute(
            get_update_properties_in_organization_mutation(fragment), payload
        )
        return result["data"]



"""Types for the user-related Kili API gateway functions."""

from dataclasses import dataclass
from typing import Optional

from kili.core.enums import OrganizationRole
from kili.domain.organization import OrganizationId
from kili.domain.user import HubspotSubscriptionStatus


@dataclass
class CreateUserDataKiliGatewayInput:
    """Input type for creating a user in Kili Gateway."""

    email: str
    firstname: Optional[str]
    lastname: Optional[str]
    password: Optional[str]
    organization_role: OrganizationRole


@dataclass
class UserDataKiliGatewayInput:
    """Input type for updating a user in Kili Gateway."""

    activated: Optional[bool] = None
    api_key: Optional[str] = None
    # auth0_id: Optional[str] = None  # refused by the backend: only used for service account  # noqa: ERA001  # pylint: disable=line-too-long
    email: Optional[str] = None
    firstname: Optional[str] = None
    has_completed_labeling_tour: Optional[bool] = None
    hubspot_subscription_status: Optional[HubspotSubscriptionStatus] = None
    lastname: Optional[str] = None
    organization: Optional[str] = None
    organization_id: Optional[OrganizationId] = None
    organization_role: Optional[OrganizationRole] = None



"""Common code for the PASCAL VOC exporter."""

import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, List, Sequence
from xml.dom import minidom

from kili.domain.ontology import JobMLTask, JobTool
from kili.services.export.exceptions import (
    NoCompatibleJobError,
    NotCompatibleInputType,
    NotCompatibleOptions,
)
from kili.services.export.format.base import AbstractExporter
from kili.services.export.media.image import get_frame_dimensions, get_image_dimensions
from kili.services.export.media.video import cut_video, get_video_dimensions
from kili.services.types import Job
from kili.utils.tqdm import tqdm


class VocExporter(AbstractExporter):
    """Common code for VOC exporter."""

    def _check_arguments_compatibility(self) -> None:
        """Check if the export label format is compatible with the export options."""
        if self.single_file:
            raise NotCompatibleOptions(
                "The Pascal VOC annotation format cannot be exported into a single file.",
            )

        if self.split_option != "merged":
            raise NotCompatibleOptions(
                "The current implementation only supports merged annotations."
            )

        if self.normalized_coordinates is True:
            raise NotCompatibleOptions(
                "The Pascal VOC annotation format does not support normalized coordinates."
            )

    def _check_project_compatibility(self) -> None:
        """Check if the export label format is compatible with the project."""
        if self.project["inputType"] not in ("IMAGE", "VIDEO"):
            raise NotCompatibleInputType(
                f"Project with input type '{self.project['inputType']}' not compatible with"
                " Pascal VOC export format."
            )

        if len(self.compatible_jobs) == 0:
            raise NoCompatibleJobError(
                f"Project needs at least one {JobMLTask.OBJECT_DETECTION} task with bounding boxes."
            )

    def _is_job_compatible(self, job: Job) -> bool:
        """Check job compatibility with the Pascal VOC format."""
        if "tools" not in job:
            return False
        return JobTool.RECTANGLE in job["tools"] and job["mlTask"] == JobMLTask.OBJECT_DETECTION

    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """Save the assets and annotations to a zip file in the Pascal VOC format."""
        self.logger.info("Exporting VOC format")

        labels_folder = self.base_folder / "labels"
        labels_folder.mkdir(parents=True, exist_ok=True)

        for asset in tqdm(assets, disable=self.disable_tqdm):
            _process_asset(asset, labels_folder, self.project["inputType"], self.compatible_jobs)

        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)


# pylint: disable=too-many-locals
def _process_asset(
    asset: Dict, labels_folder: Path, project_input_type: str, valid_jobs: Sequence[str]
) -> None:
    """Process an asset."""
    if project_input_type == "VIDEO":
        nbr_frames = len(asset.get("latestLabel", {}).get("jsonResponse", {}))
        if nbr_frames < 1:
            return
        leading_zeros = len(str(nbr_frames))

        width = height = 0
        frame_ext = ""
        # jsonContent with frames
        if isinstance(asset["jsonContent"], list) and Path(asset["jsonContent"][0]).is_file():
            width, height = get_frame_dimensions(asset)
            frame_ext = Path(asset["jsonContent"][0]).suffix

        # video with shouldUseNativeVideo set to True (no frames available)
        elif Path(asset["content"]).is_file():
            width, height = get_video_dimensions(asset)
            cut_video(asset["content"], asset, leading_zeros, Path(asset["content"]).parent)
            frame_ext = ".jpg"

        else:
            raise FileNotFoundError(f"Could not find frames or video for asset {asset}")

        for frame_id, json_response in asset["latestLabel"]["jsonResponse"].items():
            frame_name = f'{asset["externalId"]}_{str(int(frame_id)+1).zfill(leading_zeros)}'
            parameters = {"filename": f"{frame_name}{frame_ext}"}
            annotations = _convert_from_kili_to_voc_format(
                json_response, width, height, parameters, valid_jobs
            )
            filepath = labels_folder / f"{frame_name}.xml"
            filepath.parent.mkdir(parents=True, exist_ok=True)
            with open(filepath, "wb") as fout:
                fout.write(f"{annotations}\n".encode())

    elif project_input_type == "IMAGE":
        json_response = asset["latestLabel"]["jsonResponse"]
        width, height = get_image_dimensions(asset)
        filename = (
            Path(asset["content"]).name if Path(asset["content"]).is_file() else asset["externalId"]
        )
        parameters = {"filename": filename}
        annotations = _convert_from_kili_to_voc_format(
            json_response, width, height, parameters, valid_jobs
        )
        xml_filename = f'{asset["externalId"]}.xml'
        filepath = labels_folder / xml_filename
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, "wb") as fout:
            fout.write(f"{annotations}\n".encode())


def _parse_annotations(
    response: Dict,
    xml_label: ET.Element,
    img_width: int,
    img_height: int,
    valid_jobs: Sequence[str],
) -> None:
    # pylint: disable=too-many-locals
    for job_name, job_response in response.items():
        if job_name not in valid_jobs:
            continue
        if "annotations" in job_response:
            annotations = job_response["annotations"]
            for annotation in annotations:
                vertices = annotation["boundingPoly"][0]["normalizedVertices"]
                categories = annotation["categories"]
                for category in categories:
                    annotation_category = ET.SubElement(xml_label, "object")
                    name = ET.SubElement(annotation_category, "name")
                    name.text = category["name"]
                    job_name_field = ET.SubElement(annotation_category, "job_name")
                    job_name_field.text = str(job_name)
                    pose = ET.SubElement(annotation_category, "pose")
                    pose.text = "Unspecified"
                    truncated = ET.SubElement(annotation_category, "truncated")
                    truncated.text = "0"
                    difficult = ET.SubElement(annotation_category, "difficult")
                    difficult.text = "0"
                    occluded = ET.SubElement(annotation_category, "occluded")
                    occluded.text = "0"
                    bndbox = ET.SubElement(annotation_category, "bndbox")
                    x_vertices = [int(round(v["x"] * img_width)) for v in vertices]
                    y_vertices = [int(round(v["y"] * img_height)) for v in vertices]
                    xmin = ET.SubElement(bndbox, "xmin")
                    xmin.text = str(min(x_vertices))
                    xmax = ET.SubElement(bndbox, "xmax")
                    xmax.text = str(max(x_vertices))
                    ymin = ET.SubElement(bndbox, "ymin")
                    ymin.text = str(min(y_vertices))
                    ymax = ET.SubElement(bndbox, "ymax")
                    ymax.text = str(max(y_vertices))


def _provide_voc_headers(
    xml_label: ET.Element, img_width: int, img_height: int, parameters: Dict
) -> None:
    folder = ET.SubElement(xml_label, "folder")
    folder.text = parameters.get("folder", "")

    filename = ET.SubElement(xml_label, "filename")
    filename.text = parameters.get("filename", "")

    path = ET.SubElement(xml_label, "path")
    path.text = parameters.get("path", "")

    source = ET.SubElement(xml_label, "source")
    database = ET.SubElement(source, "database")
    database.text = "Kili Technology"

    size = ET.SubElement(xml_label, "size")
    width_xml = ET.SubElement(size, "width")
    width_xml.text = str(img_width)
    height_xml = ET.SubElement(size, "height")
    height_xml.text = str(img_height)
    depth = ET.SubElement(size, "depth")
    depth.text = parameters.get("depth", "3")

    segmented = ET.SubElement(xml_label, "segmented")
    segmented.text = None


def _convert_from_kili_to_voc_format(
    response: Dict, img_width: int, img_height: int, parameters: Dict, valid_jobs: Sequence[str]
) -> str:
    xml_label = ET.Element("annotation")

    _provide_voc_headers(xml_label, img_width, img_height, parameters=parameters)

    _parse_annotations(response, xml_label, img_width, img_height, valid_jobs)

    return minidom.parseString(ET.tostring(xml_label)).toprettyxml(indent="   ")



"""Asset use cases."""

import itertools
from typing import Generator, Literal, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.constants import QUERY_BATCH_SIZE
from kili.core.helpers import validate_category_search_query
from kili.core.utils.pagination import batcher
from kili.domain.asset import AssetFilters
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.services.label_data_parsing.types import Project as LabelParsingProject
from kili.use_cases.asset.asset_label_parsing import parse_labels_of_asset
from kili.use_cases.asset.media_downloader import get_download_assets_function
from kili.use_cases.base import BaseUseCases


class AssetUseCases(BaseUseCases):
    """Asset use cases."""

    # pylint: disable=too-many-arguments
    def list_assets(
        self,
        filters: AssetFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
        download_media: bool,
        local_media_dir: Optional[str],
        label_output_format: Literal["dict", "parsed_label"],
    ) -> Generator:
        """List assets with given options."""
        if filters.label_category_search:
            validate_category_search_query(filters.label_category_search)

        download_media_function, fields = get_download_assets_function(
            self._kili_api_gateway,
            download_media,
            fields,
            ProjectId(filters.project_id),
            local_media_dir,
        )
        assets_gen = self._kili_api_gateway.list_assets(filters, fields, options)

        if download_media_function is not None:
            # TODO: modify download_media function so it can take a generator of assets
            assets_lists = [
                download_media_function(assets_batch)
                for assets_batch in batcher(assets_gen, QUERY_BATCH_SIZE)
            ]
            assets_gen = (asset for asset in itertools.chain(*assets_lists))

        if label_output_format == "parsed_label":
            project = LabelParsingProject(
                **self._kili_api_gateway.get_project(
                    ProjectId(filters.project_id), ("jsonInterface", "inputType")
                )
            )
            assets_gen = (parse_labels_of_asset(asset, project) for asset in assets_gen)

        return assets_gen

    def count_assets(self, filters: AssetFilters) -> int:
        """Send a GraphQL request calling countAssets resolver."""
        if filters.label_category_search:
            validate_category_search_query(filters.label_category_search)
        return self._kili_api_gateway.count_assets(filters)



"""Base class for all client methods classes."""

from abc import ABC

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway


class BaseClientMethods(ABC):
    """Base class for all client methods classes.

    It is used to share the KiliAPIGateway between all client methods classes.

    It is not meant to be used and instantiated directly.
    """

    kili_api_gateway: KiliAPIGateway  # instantiated in the Kili client child class



"""Label import-specific errors."""


class LabelParsingError(ValueError):
    """Raised when there is a failure to parse a label."""


class MissingMetadataError(ValueError):
    """Raised when a metadata file is missing."""


class MissingTargetJobError(ValueError):
    """Raised when the target job is missing."""



"""API Key domain."""

from dataclasses import dataclass
from typing import Optional


@dataclass
class ApiKeyFilters:
    """Api key filters for running an api key search."""

    api_key_id: Optional[str] = None
    user_id: Optional[str] = None
    api_key: Optional[str] = None



"""Gets the."""

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List

from kili.adapters.http_client import HttpClient

from .exceptions import DownloadError


class AbstractContentRepository(ABC):
    """Interface to the content repository."""

    def __init__(self, router_endpoint: str, http_client: HttpClient) -> None:
        self.router_endpoint = router_endpoint
        self.http_client = http_client
        assert router_endpoint, "The router endpoint string should not be empty"

    @abstractmethod
    def get_frames(self, content_url: str) -> List[str]:
        """Get asset content frames."""

    @abstractmethod
    def get_content_stream(self, content_url: str, block_size: int) -> Iterator[Any]:
        """Get asset content stream."""

    def get_content_frames_paths(self, asset: Dict) -> List[str]:
        """Get list of links to frames from the file located at asset[jsonContent].

        Returns an empty list if `content` in the asset exists.
        """
        content_frames = []

        if not asset["content"] and asset["jsonContent"]:
            content_frames = self.get_frames(asset["jsonContent"])

        return content_frames

    def is_serving(self, url: str) -> bool:
        """Return a boolean defining if the asset is served by Kili or not."""
        return url.startswith(self.router_endpoint)


class SDKContentRepository(AbstractContentRepository):
    """Handle content fetching from the server from the SDK."""

    def get_frames(self, content_url: str) -> List[str]:
        frames: List[str] = []
        json_content_resp = self.http_client.get(content_url, timeout=30)

        if json_content_resp.ok:
            frames = list(json_content_resp.json().values())
        return frames

    def get_content_stream(self, content_url: str, block_size: int) -> Iterator[Any]:
        response = self.http_client.get(content_url, stream=True, timeout=30)
        if not response.ok:
            raise DownloadError(f"Error while downloading image {content_url}")

        return response.iter_content(block_size)



"""Formatters for projects retrieved from Kili API."""

import json
from typing import Dict

from kili.domain.types import ListOrTuple

PROJECT_JSON_FIELDS = ("jsonInterface",)


def load_project_json_fields(project: Dict, fields: ListOrTuple[str]) -> Dict:
    """Load json fields of a project."""
    if "jsonInterface" in fields and isinstance(project.get("jsonInterface"), str):
        try:
            project["jsonInterface"] = json.loads(project["jsonInterface"])
        except json.JSONDecodeError:
            project["jsonInterface"] = {}
    return project



"""Collection of notification's related GraphQL queries and mutations."""

GQL_COUNT_NOTIFICATIONS = """
    query countNotifications($where: NotificationWhere!) {
        data: countNotifications(where: $where)
    }
    """


def get_notifications_query(fragment: str) -> str:
    """Get the query for notifications."""
    return f"""
        query notifications($where: NotificationWhere!, $first: PageSize!, $skip: Int!) {{
            data: notifications(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """



"""Mixin extending Kili API Gateway class with Cloud Storage related operations."""

from typing import Dict, Generator, Optional

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.cloud_storage.types import DataIntegrationData
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.domain.cloud_storage import (
    DataConnectionFilters,
    DataConnectionId,
    DataDifferenceType,
    DataIntegrationFilters,
    DataIntegrationId,
)
from kili.domain.types import ListOrTuple

from .mappers import (
    add_data_connection_data_mapper,
    compute_data_connection_difference_data_mapper,
    data_connection_where_mapper,
    data_integration_where_mapper,
    integration_data_mapper,
)
from .operations import (
    GQL_COUNT_DATA_INTEGRATIONS,
    GQL_DELETE_DATA_INTEGRATION,
    get_add_data_connection_mutation,
    get_compute_data_connection_differences_mutation,
    get_create_integration_mutation,
    get_data_connection_query,
    get_list_data_connections_query,
    get_list_data_integrations_query,
    get_update_integration_mutation,
    get_validate_data_connection_differences_mutation,
)
from .types import (
    AddDataConnectionKiliAPIGatewayInput,
    DataConnectionComputeDifferencesKiliAPIGatewayInput,
)


class CloudStorageOperationMixin(BaseOperationMixin):
    """GraphQL Mixin extending GraphQL Gateway class with Cloud Storage related operations."""

    def list_data_integrations(
        self,
        data_integration_filters: DataIntegrationFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List data integrations."""
        fragment = fragment_builder(fields)
        query = get_list_data_integrations_query(fragment)
        where = data_integration_where_mapper(filters=data_integration_filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving data integrations", GQL_COUNT_DATA_INTEGRATIONS
        )

    def count_data_integrations(self, data_integration_filters: DataIntegrationFilters) -> int:
        """Count data integrations."""
        where = data_integration_where_mapper(filters=data_integration_filters)
        variables = {"where": where}
        result = self.graphql_client.execute(GQL_COUNT_DATA_INTEGRATIONS, variables)
        return result["data"]

    def list_data_connections(
        self,
        data_connection_filters: DataConnectionFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List data connections."""
        fragment = fragment_builder(fields)
        query = get_list_data_connections_query(fragment)
        where = data_connection_where_mapper(filters=data_connection_filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving data connections", GQL_COUNT_DATA_INTEGRATIONS
        )

    def get_data_connection(
        self, data_connection_id: DataConnectionId, fields: ListOrTuple[str]
    ) -> Dict:
        """Get data connection."""
        fragment = fragment_builder(fields)
        query = get_data_connection_query(fragment)
        result = self.graphql_client.execute(
            query=query, variables={"where": {"id": data_connection_id}}
        )
        return result["data"]

    def add_data_connection(
        self, data: AddDataConnectionKiliAPIGatewayInput, fields: ListOrTuple[str]
    ) -> Dict:
        """Add data connection to a project."""
        fragment = fragment_builder(fields)
        query = get_add_data_connection_mutation(fragment)
        variables = {"data": add_data_connection_data_mapper(data)}
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def compute_data_connection_differences(
        self,
        data_connection_id: DataConnectionId,
        data: Optional[DataConnectionComputeDifferencesKiliAPIGatewayInput],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Compute data connection differences."""
        fragment = fragment_builder(fields)
        query = get_compute_data_connection_differences_mutation(fragment)
        variables = {"where": {"id": data_connection_id}}
        if data is not None:
            variables["data"] = compute_data_connection_difference_data_mapper(data)
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def validate_data_connection_differences(
        self,
        data_connection_id: DataConnectionId,
        data_difference_type: DataDifferenceType,
        fields: ListOrTuple[str],
    ) -> Dict:
        """Validate data connection differences."""
        fragment = fragment_builder(fields)
        query = get_validate_data_connection_differences_mutation(fragment)
        variables = {"where": {"connectionId": data_connection_id, "type": data_difference_type}}
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def create_data_integration(self, data: DataIntegrationData, fields: ListOrTuple[str]) -> Dict:
        """Create a data integration."""
        fragment = fragment_builder(fields)
        query = get_create_integration_mutation(fragment)
        variables = {"data": integration_data_mapper(data)}
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def update_data_integration(
        self,
        data_integration_id: DataIntegrationId,
        data_integration_data: DataIntegrationData,
        fields: ListOrTuple[str],
    ) -> Dict:
        """Update a data integration."""
        fragment = fragment_builder(fields)
        query = get_update_integration_mutation(fragment)
        variables = {
            "data": integration_data_mapper(data=data_integration_data),
            "where": {"id": data_integration_id},
        }
        result = self.graphql_client.execute(query, variables)
        return result["data"]

    def delete_data_integration(self, data_integration_id: DataIntegrationId) -> str:
        """Delete a data integration."""
        variables = {"where": {"id": data_integration_id}}
        result = self.graphql_client.execute(GQL_DELETE_DATA_INTEGRATION, variables)
        return result["data"]



"""Cloud storage use cases."""

import logging
import time
from datetime import datetime
from logging import Logger
from typing import Dict, Generator, List, Optional

from tenacity import Retrying
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_exponential

from kili.adapters.kili_api_gateway.cloud_storage.mappers import integration_data_mapper
from kili.adapters.kili_api_gateway.cloud_storage.types import (
    AddDataConnectionKiliAPIGatewayInput,
    DataConnectionComputeDifferencesKiliAPIGatewayInput,
    DataIntegrationData,
)
from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.asset import AssetFilters
from kili.domain.cloud_storage import (
    DataConnectionFilters,
    DataConnectionId,
    DataDifferenceType,
    DataIntegrationFilters,
    DataIntegrationId,
    DataIntegrationPlatform,
    DataIntegrationStatus,
    ProjectId,
)
from kili.domain.organization import OrganizationId
from kili.domain.types import ListOrTuple
from kili.use_cases.base import BaseUseCases


class CloudStorageUseCases(BaseUseCases):
    """CloudStorage use cases."""

    def list_data_connections(
        self,
        data_connection_filters: DataConnectionFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List data connections."""
        return self._kili_api_gateway.list_data_connections(
            data_connection_filters=data_connection_filters, fields=fields, options=options
        )

    def get_data_connection(
        self, data_connection_id: DataConnectionId, fields: ListOrTuple[str]
    ) -> Dict:
        """Get data connection."""
        return self._kili_api_gateway.get_data_connection(
            data_connection_id=data_connection_id, fields=fields
        )

    def list_data_integrations(
        self,
        data_integration_filters: DataIntegrationFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List data integrations."""
        return self._kili_api_gateway.list_data_integrations(
            data_integration_filters=data_integration_filters, fields=fields, options=options
        )

    def count_data_integrations(self, data_integration_filters: DataIntegrationFilters) -> int:
        """Count data integrations."""
        return self._kili_api_gateway.count_data_integrations(data_integration_filters)

    def add_data_connection(
        self,
        data_integration_id: DataIntegrationId,
        project_id: ProjectId,
        selected_folders: Optional[List[str]],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Add data connection to a project."""
        if (
            self._kili_api_gateway.count_data_integrations(
                DataIntegrationFilters(id=data_integration_id)
            )
            == 0
        ):
            raise ValueError(f"Cloud storage integration with id {data_integration_id} not found.")
        return self._kili_api_gateway.add_data_connection(
            fields=fields,
            data=AddDataConnectionKiliAPIGatewayInput(
                is_checking=False,
                integration_id=data_integration_id,
                last_checked=datetime.now(),
                project_id=project_id,
                selected_folders=selected_folders,
            ),
        )

    def synchronize_data_connection(
        self,
        data_connection_id: DataConnectionId,
        delete_extraneous_files: bool,
        dry_run: bool,
        logger: Optional[Logger],
    ) -> None:
        """Synchronize data connection."""
        if logger is None:
            logger = logging.getLogger(__name__)
            logger.addHandler(logging.NullHandler())

        logger.info("Synchronizing data connection: %s", data_connection_id)

        _compute_differences(data_connection_id, self._kili_api_gateway)

        data_connection = self._kili_api_gateway.get_data_connection(
            data_connection_id,
            fields=(
                "id",
                "dataDifferencesSummary.added",
                "dataDifferencesSummary.removed",
                "dataDifferencesSummary.total",
                "isChecking",
                "numberOfAssets",
                "projectId",
            ),
        )
        if data_connection["isChecking"]:
            raise ValueError(f"Data connection should not be checking: {data_connection}")

        added = data_connection["dataDifferencesSummary"]["added"]
        removed = data_connection["dataDifferencesSummary"]["removed"]
        total = data_connection["dataDifferencesSummary"]["total"]
        nb_assets = data_connection["numberOfAssets"]

        logger.info("Currently %d asset(s) imported from the data connection.", nb_assets)

        if total == 0:
            logger.info("No differences found. Nothing to synchronize.")
            return

        logger.info(
            "Found %d difference(s): %d asset(s) to add, %d asset(s) to remove.",
            total,
            added,
            removed,
        )

        if dry_run:
            # pylint: disable=unnecessary-lambda-assignment
            validate_data_differences_func = lambda *args, **kwargs: None  # noqa: ARG005
            logger.info("Dry run: no data will be added or removed.")
        else:
            validate_data_differences_func = _validate_data_differences

        if added > 0:
            validate_data_differences_func(
                data_connection_id=data_connection_id,
                data_difference_type=DataDifferenceType.ADD,
                kili_api_gateway=self._kili_api_gateway,
            )
            logger.info("Added %d file(s).", added)

        if removed > 0:
            if delete_extraneous_files:
                validate_data_differences_func(
                    data_connection_id=data_connection_id,
                    data_difference_type=DataDifferenceType.REMOVE,
                    kili_api_gateway=self._kili_api_gateway,
                )
                logger.info("Removed %d extraneous file(s).", removed)
            else:
                logger.info(
                    "Use delete_extraneous_files=True to remove %d extraneous file(s).",
                    removed,
                )

    def create_data_integration(
        self,
        allowed_paths: Optional[List[str]],
        allowed_projects: Optional[List[str]],
        aws_access_point_arn: Optional[str],
        aws_role_arn: Optional[str],
        aws_role_external_id: Optional[str],
        azure_connection_url: Optional[str],
        azure_is_using_service_credentials: Optional[bool],
        azure_sas_token: Optional[str],
        azure_tenant_id: Optional[str],
        gcp_bucket_name: Optional[str],
        include_root_files: Optional[str],
        internal_processing_authorized: Optional[str],
        name: str,
        platform: DataIntegrationPlatform,
        s3_access_key: Optional[str],
        s3_bucket_name: Optional[str],
        s3_endpoint: Optional[str],
        s3_region: Optional[str],
        s3_secret_key: Optional[str],
        s3_session_token: Optional[str],
        fields: ListOrTuple[str],
    ) -> Dict:
        """Create a data integration."""
        data = DataIntegrationData(
            allowed_paths=allowed_paths,
            allowed_projects=allowed_projects,
            aws_access_point_arn=aws_access_point_arn,
            aws_role_arn=aws_role_arn,
            aws_role_external_id=aws_role_external_id,
            azure_connection_url=azure_connection_url,
            azure_is_using_service_credentials=azure_is_using_service_credentials,
            azure_sas_token=azure_sas_token,
            azure_tenant_id=azure_tenant_id,
            gcp_bucket_name=gcp_bucket_name,
            include_root_files=include_root_files,
            internal_processing_authorized=internal_processing_authorized,
            name=name,
            organization_id=OrganizationId(""),
            platform=platform,
            status="CHECKING",
            s3_access_key=s3_access_key,
            s3_bucket_name=s3_bucket_name,
            s3_endpoint=s3_endpoint,
            s3_region=s3_region,
            s3_secret_key=s3_secret_key,
            s3_session_token=s3_session_token,
        )

        return self._kili_api_gateway.create_data_integration(fields=fields, data=data)

    def update_data_integration(
        self,
        data_integration_id: DataIntegrationId,
        allowed_paths: Optional[List[str]],
        allowed_projects: Optional[List[str]],
        aws_access_point_arn: Optional[str],
        aws_role_arn: Optional[str],
        aws_role_external_id: Optional[str],
        azure_connection_url: Optional[str],
        azure_is_using_service_credentials: Optional[bool],
        azure_sas_token: Optional[str],
        azure_tenant_id: Optional[str],
        gcp_bucket_name: Optional[str],
        include_root_files: Optional[str],
        internal_processing_authorized: Optional[str],
        name: Optional[str],
        platform: Optional[DataIntegrationPlatform],
        organization_id: Optional[str],
        s3_access_key: Optional[str],
        s3_bucket_name: Optional[str],
        s3_endpoint: Optional[str],
        s3_region: Optional[str],
        s3_secret_key: Optional[str],
        s3_session_token: Optional[str],
        status: Optional[DataIntegrationStatus],
    ) -> Dict:
        """Update data integration."""
        organization_id = OrganizationId(organization_id) if organization_id else None
        data = DataIntegrationData(
            allowed_paths=allowed_paths,
            allowed_projects=allowed_projects,
            aws_access_point_arn=aws_access_point_arn,
            aws_role_arn=aws_role_arn,
            aws_role_external_id=aws_role_external_id,
            azure_connection_url=azure_connection_url,
            azure_is_using_service_credentials=azure_is_using_service_credentials,
            azure_sas_token=azure_sas_token,
            azure_tenant_id=azure_tenant_id,
            gcp_bucket_name=gcp_bucket_name,
            include_root_files=include_root_files,
            internal_processing_authorized=internal_processing_authorized,
            name=name,
            organization_id=organization_id,
            platform=platform,
            status=status,
            s3_access_key=s3_access_key,
            s3_bucket_name=s3_bucket_name,
            s3_endpoint=s3_endpoint,
            s3_region=s3_region,
            s3_secret_key=s3_secret_key,
            s3_session_token=s3_session_token,
        )
        fields = tuple(
            name for name, val in integration_data_mapper(data).items() if val is not None
        )
        if "id" not in fields:
            fields += ("id",)

        return self._kili_api_gateway.update_data_integration(
            data_integration_id,
            data_integration_data=data,
            fields=fields,
        )

    def delete_data_integration(self, data_integration_id: DataIntegrationId) -> str:
        """Delete a data integration."""
        return self._kili_api_gateway.delete_data_integration(data_integration_id)


def _compute_differences(
    data_connection_id: DataConnectionId, kili_api_gateway: KiliAPIGateway
) -> None:
    """Compute data connection differences.

    Calls the compute_data_connection_differences mutation
    and waits until the data connection is not checking anymore.
    """
    data_connection = kili_api_gateway.get_data_connection(
        data_connection_id,
        fields=(
            "dataIntegration.azureIsUsingServiceCredentials",
            "dataIntegration.platform",
            "dataIntegration.azureSASToken",
            "dataIntegration.azureConnectionURL",
            "dataIntegration.id",
            "selectedFolders",
            "projectId",
            "isChecking",
            "exclude",
            "include",
            "prefix",
        ),
    )
    data_integration = data_connection["dataIntegration"]

    if data_connection["isChecking"]:
        raise ValueError(
            f"Cannot compute differences for data connection {data_connection_id}."
            " Data connection is already currently checking."
        )

    blob_paths = warnings = content_types = None

    # for azure using credentials, it is required to provide the blob paths to compute the diffs
    if (
        data_integration["platform"] == "Azure"
        and data_integration["azureIsUsingServiceCredentials"]
    ):
        try:
            # pylint: disable=import-outside-toplevel
            from .azure import (
                get_blob_paths_azure_data_connection_with_service_credentials,
            )
        except ImportError as err:
            raise ImportError(
                "The azure-storage-blob package is required to use Azure buckets. "
                " Run `pip install kili[azure]` to install it."
            ) from err

        project_input_type = kili_api_gateway.get_project(
            project_id=data_connection["projectId"], fields=("inputType",)
        )["inputType"]
        (
            blob_paths,
            warnings,
            content_types,
        ) = get_blob_paths_azure_data_connection_with_service_credentials(
            input_type=project_input_type,
            data_connection=data_connection,
            data_integration=data_integration,
        )

    kili_api_gateway.compute_data_connection_differences(
        data_connection_id=data_connection_id,
        fields=("id",),
        data=(
            DataConnectionComputeDifferencesKiliAPIGatewayInput(
                blob_paths=blob_paths, warnings=warnings, content_types=content_types
            )
            if blob_paths is not None and warnings is not None and content_types is not None
            else None
        ),
    )
    time.sleep(1)  # backend needs some time to refresh the "isChecking"

    for attempt in Retrying(
        wait=wait_exponential(multiplier=1, min=1, max=4),
        stop=stop_after_delay(5 * 60),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    ):
        with attempt:
            data_connection = kili_api_gateway.get_data_connection(
                data_connection_id, fields=("isChecking",)
            )
            if data_connection["isChecking"]:
                raise ValueError(f"Data connection is still checking: {data_connection}")

    time.sleep(1)  # backend needs some time to refresh the "isChecking"


def _validate_data_differences(
    data_difference_type: DataDifferenceType,
    data_connection_id: DataConnectionId,
    kili_api_gateway: KiliAPIGateway,
) -> None:
    """Validate data differences for a data connection.

    Calls the validate_data_connection_differences mutation
    and waits until the number of assets in the project is correct.
    """
    data_connection = kili_api_gateway.get_data_connection(
        data_connection_id,
        fields=(
            "dataDifferencesSummary.added",
            "dataDifferencesSummary.removed",
            "dataDifferencesSummary.total",
            "projectId",
        ),
    )
    asset_diff: int = data_connection["dataDifferencesSummary"][
        "added" if data_difference_type == DataDifferenceType.ADD else "removed"
    ]

    filters = AssetFilters(project_id=data_connection["projectId"])
    nb_assets_before = kili_api_gateway.count_assets(filters)

    kili_api_gateway.validate_data_connection_differences(
        data_connection_id=data_connection_id,
        data_difference_type=data_difference_type,
        fields=("id",),
    )

    for attempt in Retrying(
        wait=wait_exponential(multiplier=1, min=1, max=4),
        stop=stop_after_delay(5 * 60),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    ):
        with attempt:
            nb_assets_after = kili_api_gateway.count_assets(filters)
            if abs(nb_assets_after - nb_assets_before) != asset_diff:
                raise ValueError(
                    "Number of assets in project after validation is not correct: before:"
                    f" {nb_assets_before} assets, after: {nb_assets_after} assets,"
                    f" dataDifferencesSummary: {asset_diff}"
                )



"""Point label utils."""

from typing import Any, Dict, List, Optional


def kili_point_to_geojson_point(point: Dict[str, float]) -> Dict[str, Any]:
    """Convert a Kili point to a geojson point.

    Args:
        point: a Kili point (vertex).

    Returns:
        A geojson point.

    !!! Example
        ```python
        >>> point = {"x": 1.0, "y": 2.0}
        >>> kili_point_to_geojson_point(point)
        {
            "type": "Point",
            "coordinates": [1.0, 2.0]
        }
        ```
    """
    return {"type": "Point", "coordinates": [point["x"], point["y"]]}


def kili_point_annotation_to_geojson_point_feature(
    point_annotation: Dict[str, Any], job_name: Optional[str] = None
) -> Dict[str, Any]:
    """Convert a Kili point annotation to a geojson point feature.

    Args:
        point_annotation: a Kili point annotation.
        job_name: the name of the job to which the annotation belongs.

    Returns:
        A geojson point feature.

    !!! Example
        ```python
        >>> point = {
            'children': {},
            'point': {'x': -79.0, 'y': -3.0},
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'marker'
        }
        >>> kili_point_annotation_to_geojson_point_feature(point)
        {
            'type': 'Feature',
            'geometry': {
                'type': 'Point',
                'coordinates': [-79.0, -3.0]},
                'id': 'mid_object',
                'properties': {
                    'kili': {
                        'categories': [{'name': 'A'}],
                        'children': {},
                        'type': 'marker'
                    }
                }
            }
        }
        ```
    """
    point = point_annotation
    assert point["type"] == "marker", f"Annotation type must be `marker`, got: {point['type']}"

    ret = {"type": "Feature", "geometry": kili_point_to_geojson_point(point["point"])}
    if "mid" in point:
        ret["id"] = point["mid"]
    ret["properties"] = {"kili": {k: v for k, v in point.items() if k not in ["point", "mid"]}}
    if job_name is not None:
        ret["properties"]["kili"]["job"] = job_name
    return ret


def geojson_point_feature_to_kili_point_annotation(
    point: Dict[str, Any],
    categories: Optional[List[Dict]] = None,
    children: Optional[Dict] = None,
    mid: Optional[str] = None,
) -> Dict[str, Any]:
    # pylint: disable=line-too-long
    """Convert a geojson point feature to a Kili point annotation.

    Args:
        point: a geojson point feature.
        categories: the categories of the annotation.
            If not provided, the categories are taken from the `kili` key of the geojson feature properties.
        children: the children of the annotation.
            If not provided, the children are taken from the `kili` key of the geojson feature properties.
        mid: the mid of the annotation.
            If not provided, the mid is taken from the `id` key of the geojson feature.

    Returns:
        A Kili point annotation.

    !!! Example
        ```python
        >>> point = {
            'type': 'Feature',
            'geometry': {'type': 'Point', 'coordinates': [-79.0, -3.0]},
            'id': 'mid_object',
            'properties': {'kili': {'categories': [{'name': 'A'}]}}
        }
        >>> geojson_point_feature_to_kili_point_annotation(point)
        {
            'children': {},
            'point': {'x': -79.0, 'y': -3.0},
            'categories': [{'name': 'A'}],
            'mid': 'mid_object',
            'type': 'marker'
        }
        ```
    """
    assert point.get("type") == "Feature", f"Feature type must be `Feature`, got: {point['type']}"
    assert (
        point["geometry"]["type"] == "Point"
    ), f"Geometry type must be `Point`, got: {point['geometry']['type']}"

    children = children or point["properties"].get("kili", {}).get("children", {})
    categories = categories or point["properties"]["kili"]["categories"]

    ret = {
        "children": children,
        "categories": categories,
        "type": "marker",
    }
    ret["point"] = {
        "x": point["geometry"]["coordinates"][0],
        "y": point["geometry"]["coordinates"][1],
    }

    if mid is not None:
        ret["mid"] = str(mid)
    elif "id" in point:
        ret["mid"] = str(point["id"])

    return ret



"""GraphQL User operations."""


def get_users_query(fragment: str) -> str:
    """Return the GraphQL users query."""
    return f"""
        query users($where: UserWhere!, $first: PageSize!, $skip: Int!) {{
            data: users(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """


def get_current_user_query(fragment: str) -> str:
    """Return the GraphQL current user query."""
    return f"""
      query me {{
          data: me {{
              {fragment}
          }}
      }}
    """


def get_create_user_mutation(fragment: str) -> str:
    """Return the GraphQL create user mutation."""
    return f"""
      mutation(
          $data: CreateUserData!
      ) {{
        data: createUser(
            data: $data
        ) {{
          {fragment}
        }}
      }}
    """


def get_update_password_mutation(fragment: str) -> str:
    """Return the GraphQL update password mutation."""
    return f"""
mutation(
    $data: UpdatePasswordData!
    $where: UserWhere!
) {{
  data: updatePassword(
    data: $data
    where: $where
  ) {{
    {fragment}
  }}
}}
"""


def get_update_user_mutation(fragment: str) -> str:
    """Return the GraphQL update user mutation."""
    return f"""
      mutation updatePropertiesInUser( $data: UserData!, $where: UserWhere!) {{
        data: updatePropertiesInUser( data: $data,  where: $where) {{
          {fragment}
        }}
      }}
      """


GQL_COUNT_USERS = """
    query countUsers($where: UserWhere!) {
        data: countUsers(where: $where)
    }
    """


def get_reset_password_mutation(fragment: str) -> str:
    """Return the GraphQL reset password mutation."""
    return f"""
    mutation($where: UserWhere!) {{
      data: resetPassword(where: $where) {{
        {fragment}
      }}
    }}
"""



"""GraphQL payload data mappers for asset operations."""

import warnings

from kili.domain.asset import AssetFilters

MAX_PARTIAL_EXTERNAL_IDS_FILTER = 10


def asset_where_mapper(filters: AssetFilters):
    """Build the GraphQL AssetWhere variable to be sent in an operation."""
    if (
        filters.external_id_in is not None
        and len(filters.external_id_in) > MAX_PARTIAL_EXTERNAL_IDS_FILTER
    ):
        warnings.warn(
            f"Requesting more than {MAX_PARTIAL_EXTERNAL_IDS_FILTER} partial external IDs"
            "in a single query is deprecated. You can use the `external_id_strictly_in`"
            f"field to filter by more than {MAX_PARTIAL_EXTERNAL_IDS_FILTER} external IDs. "
            "This limit will be enforced in next versions.",
            stacklevel=5,
        )
    return {
        "id": filters.asset_id,
        "project": {
            "id": filters.project_id,
        },
        "externalIdStrictlyIn": filters.external_id_strictly_in,
        "externalIdIn": filters.external_id_in,
        "statusIn": filters.status_in,
        "consensusMarkGte": filters.consensus_mark_gte,
        "consensusMarkLte": filters.consensus_mark_lte,
        "honeypotMarkGte": filters.honeypot_mark_gte,
        "honeypotMarkLte": filters.honeypot_mark_lte,
        "idIn": filters.asset_id_in,
        "idNotIn": filters.asset_id_not_in,
        "metadata": filters.metadata_where,
        "label": {
            "typeIn": filters.label_type_in,
            "authorIn": filters.label_author_in,
            "consensusMarkGte": filters.label_consensus_mark_gte,
            "consensusMarkLte": filters.label_consensus_mark_lte,
            "createdAt": filters.label_created_at,
            "createdAtGte": filters.label_created_at_gte,
            "createdAtLte": filters.label_created_at_lte,
            "honeypotMarkGte": filters.label_honeypot_mark_gte,
            "honeypotMarkLte": filters.label_honeypot_mark_lte,
            "search": filters.label_category_search,
            "reviewerIn": filters.label_reviewer_in,
        },
        "skipped": filters.skipped,
        "updatedAtGte": filters.updated_at_gte,
        "updatedAtLte": filters.updated_at_lte,
        "createdAtGte": filters.created_at_gte,
        "createdAtLte": filters.created_at_lte,
        "inferenceMarkGte": filters.inference_mark_gte,
        "inferenceMarkLte": filters.inference_mark_lte,
        "issue": {
            "type": filters.issue_type,
            "status": filters.issue_status,
        },
    }



"""Helpers for the asset domain."""

from typing import Optional

from kili.domain.asset import AssetExternalId, AssetId
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.exceptions import IncompatibleArgumentsError, MissingArgumentError


def check_asset_identifier_arguments(
    project_id: Optional[ProjectId],
    asset_id_array: Optional[ListOrTuple[AssetId]],
    asset_external_id_array: Optional[ListOrTuple[AssetExternalId]],
) -> None:
    # pylint: disable=line-too-long
    """Check that a list of assets can be identified either by their asset IDs or their external IDs."""
    if asset_id_array is not None:
        if asset_external_id_array is not None:
            raise IncompatibleArgumentsError(
                "Either provide asset IDs or asset external IDs. Not both at the same time."
            )
        return

    if project_id is None or asset_external_id_array is None:
        raise MissingArgumentError(
            "Either provide asset IDs or project ID with asset external IDs."
        )



"""Mixin extending Kili API Gateway class with Asset related operations."""

from typing import Dict, Generator, List

from kili.adapters.kili_api_gateway.asset.formatters import (
    load_asset_json_fields,
)
from kili.adapters.kili_api_gateway.asset.mappers import asset_where_mapper
from kili.adapters.kili_api_gateway.asset.operations import (
    GQL_COUNT_ASSETS,
    GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS,
    get_assets_query,
)
from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.adapters.kili_api_gateway.label.annotation_to_json_response import (
    AnnotationsToJsonResponseConverter,
)
from kili.adapters.kili_api_gateway.label.common import get_annotation_fragment
from kili.adapters.kili_api_gateway.project.common import get_project
from kili.domain.asset import AssetFilters
from kili.domain.types import ListOrTuple


class AssetOperationMixin(BaseOperationMixin):
    """Mixin extending Kili API Gateway class with Assets related operations."""

    def list_assets(
        self,
        filters: AssetFilters,
        fields: ListOrTuple[str],
        options: QueryOptions,
    ) -> Generator[Dict, None, None]:
        """List assets with given options."""
        if "labels.jsonResponse" in fields or "latestLabel.jsonResponse" in fields:
            # Check if we can get the jsonResponse of if we need to rebuild it.
            project_info = get_project(
                self.graphql_client, filters.project_id, ("inputType", "jsonInterface")
            )
            if project_info["inputType"] in {"VIDEO", "LLM_RLHF"}:
                yield from self.list_assets_split(filters, fields, options, project_info)
                return

        fragment = fragment_builder(fields)
        query = get_assets_query(fragment)
        where = asset_where_mapper(filters)
        assets_gen = PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query,
            where,
            options,
            "Retrieving assets",
            GQL_COUNT_ASSETS,
            "id" if "id" in fields else None,
        )
        assets_gen = (
            load_asset_json_fields(asset, fields, self.http_client) for asset in assets_gen
        )

        yield from assets_gen

    def list_assets_split(
        self, filters: AssetFilters, fields: ListOrTuple[str], options: QueryOptions, project_info
    ) -> Generator[Dict, None, None]:
        """List assets with given options."""
        options = QueryOptions(
            options.disable_tqdm,
            options.first,
            options.skip,
            min(options.batch_size, 10 if project_info["inputType"] == "VIDEO" else 50),
        )

        inner_annotation_fragment = get_annotation_fragment()
        annotation_fragment = f"""
            annotations {{
                {inner_annotation_fragment}
            }}
        """
        fragment = fragment_builder(
            fields, {"labels": annotation_fragment, "latestLabel": annotation_fragment}
        )
        query = get_assets_query(fragment)
        where = asset_where_mapper(filters)
        assets_gen = PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving assets", GQL_COUNT_ASSETS
        )
        assets_gen = (
            load_asset_json_fields(asset, fields, self.http_client) for asset in assets_gen
        )

        converter = AnnotationsToJsonResponseConverter(
            json_interface=project_info["jsonInterface"],
            project_input_type=project_info["inputType"],
        )
        for asset in assets_gen:
            if "latestLabel.jsonResponse" in fields and asset.get("latestLabel"):
                converter.patch_label_json_response(
                    asset["latestLabel"], asset["latestLabel"]["annotations"]
                )
                if "latestLabel.annotations" not in fields:
                    asset["latestLabel"].pop("annotations")

            if "labels.jsonResponse" in fields:
                for label in asset.get("labels", []):
                    converter.patch_label_json_response(label, label["annotations"])
                    if "labels.annotations" not in fields:
                        label.pop("annotations")
            yield asset

    def count_assets(self, filters: AssetFilters) -> int:
        """Send a GraphQL request calling countIssues resolver."""
        where = asset_where_mapper(filters)
        payload = {"where": where}
        count_result = self.graphql_client.execute(GQL_COUNT_ASSETS, payload)
        count: int = count_result["data"]
        return count

    def create_upload_bucket_signed_urls(self, file_paths: List[str]) -> List[str]:
        """Send a GraphQL request calling createUploadBucketSignedUrls resolver."""
        payload = {
            "filePaths": file_paths,
        }
        urls_response = self.graphql_client.execute(GQL_CREATE_UPLOAD_BUCKET_SIGNED_URLS, payload)
        return urls_response["urls"]
